<?php
// Autor (c) Miroslav Novak, www.platiti.cz
// Pouzivani bez souhlasu autora neni povoleno
// #Ver:PRV108-60-g39033a71:2024-08-14#

namespace Platiti\GPWebPayGpeNs\UniModul;

use Exception;
use Throwable;
use stdClass;

class WebTools {

	/** @var UniLogger */
	private $logger;


	/**
	 * @param UniLogger $logger
	 * @return void
	 */
	function __construct($logger) {
		$this->logger = $logger;
	}


	/**
	 * @param string $method
	 * @param string $url
	 * @param mixed $post_data
	 * @param ?array<string, string> $headers
	 * @param ?array<string, mixed> $additional_parameters suppress_logging=false, timeout, disable_ssl_verification=false, follow_redirects=true
	 * @return mixed
	 */
	public function httpRequestJson($method, $url, $post_data=null, $headers=[], $additional_parameters=[]) {
		return $this->json_decode($this->httpRequest($method, $url, $this->json_encode($post_data), $headers, $additional_parameters));
	}


	/**
	 * @param string $method
	 * @param string $url
	 * @param ?string $post_data
	 * @param ?array<string, string> $headers
	 * @param ?array<string, mixed> $additional_parameters suppress_logging=false, timeout, disable_ssl_verification=false, follow_redirects=true
	 * @return string
	 */
	public function httpRequest($method, $url, $post_data=null, $headers=[], $additional_parameters=[]) {
		$get_additional_parameter = function ($name, $default) use ($additional_parameters) {
			return isset($additional_parameters[$name]) ? $additional_parameters[$name] : $default;
		};
		$suppress_logging = $get_additional_parameter('suppress_logging', false);

		$ch = curl_init();

		if (!$suppress_logging)
			$this->logger->writeLogNoNewLines("INFO UniModul: REQUEST - Sending request: METHOD: {$method} URL: {$url}".(isset($post_data) ? " PostData: '$post_data'" : "") . (!empty($headers) ? " Headers: ".print_r($headers, true) : ""));

		curl_setopt($ch, CURLOPT_URL, $url);
		curl_setopt($ch, CURLOPT_FOLLOWLOCATION, $get_additional_parameter('follow_redirects', true));
		curl_setopt($ch, CURLOPT_HEADER, false);

		if (!empty($headers)) {
			$curl_headers = [];
			foreach ($headers as $key => $value) {
				$curl_headers[] = "$key: $value";
			}
			curl_setopt($ch, CURLOPT_HTTPHEADER, $curl_headers);
		}

		if ($method==="POST" || $method==="PUT") {
			curl_setopt($ch, CURLOPT_CUSTOMREQUEST, $method);
			curl_setopt($ch, CURLOPT_POSTFIELDS, $post_data);
		}
		// total vypnuti kontroly https certifikatu, NEBEZPECNE!
		if (!empty($GLOBALS['curlopt_insecure']) || $get_additional_parameter('disable_ssl_verification', false)) {
			$this->logger->writeLog("WARN UniModul: REQUEST - SSL verification disabled!");
			curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);
			curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, false);
		}
		if (isset($additional_parameters['timeout'])) {
			curl_setopt($ch, CURLOPT_TIMEOUT, $additional_parameters['timeout']);
		}
		curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
		$data = curl_exec($ch);
		if ($data === true) throw new Exception('curl_exec didnt return transfer'); // should never execute

		$curlErrno = curl_errno($ch);
		$curlError = curl_error($ch);
		$httpStatusCode = curl_getinfo($ch, CURLINFO_HTTP_CODE);
		$contentType = curl_getinfo($ch,  CURLINFO_CONTENT_TYPE);
		curl_close($ch);

		if ($curlErrno || $data === false) {
			if (!$suppress_logging)
				$this->logger->writeLogNoNewLines("ERROR UniModul: RESPONSE - Curl returned an error: ErrNo: {$curlErrno} Error: {$curlError}");
			throw new Exception("Curl returned an error. Error: " . $curlError);
		}
		if ($httpStatusCode < 200 || $httpStatusCode >= 300) {
			if (!$suppress_logging)
				$this->logger->writeLogNoNewLines("ERROR UniModul: RESPONSE - HTTP returned an error: StatusCode: {$httpStatusCode} ContentType: {$contentType} Response: {$data}");
			throw new HttpRequestException($httpStatusCode, $data);
		}
		if (!$suppress_logging)
			$this->logger->writeLogNoNewLines("INFO UniModul: RESPONSE - Finished: StatusCode: {$httpStatusCode} Response: {$data}");
		return $data;
	}


	/**
	 * @param string $oldUrl
	 * @param array<string, string> $argsArray
	 * @return string
	 */
	public function appendUrlQueryArgs($oldUrl, $argsArray) {
		if (empty($argsArray)) return $oldUrl;
		return $oldUrl . (strpos($oldUrl, '?')===false ? '?' : '&') . http_build_query($argsArray);
	}



	/**
	 * Decodes json and throws exception on error
	 * For PHP since 7.3.0 the JSON_THROW_ON_ERROR flag in php function json_decode could be used
	 * @param string $s
	 * @return mixed
	 */
	public function json_decode($s) {
		$r = json_decode($s);
		if (json_last_error() !== JSON_ERROR_NONE) {
			throw new Exception("json_decode " . json_last_error_msg());
		}
		return $r;
	}

	/**
	 * Encodes json and throws exception on error
	 * For PHP since 7.3.0 the JSON_THROW_ON_ERROR flag in php function json_decode could be used
	 * @param mixed $a
	 * @return string
	 */
	public function json_encode($a) {
		$serialize_precision = ini_set('serialize_precision', 10); // Avoiding values e.g. "amount": 1390.8999999999999 and errors such as Field=amount has an error message=numeric value out of bounds (<6 digits>.<2 digits> expected)
		$precision = ini_set('precision', 10);
		$s = json_encode($a, JSON_PRETTY_PRINT | JSON_UNESCAPED_SLASHES | JSON_UNESCAPED_UNICODE);
		ini_set('serialize_precision', $serialize_precision);
		ini_set('precision', $precision);
		if ($s === false) {
			throw new Exception("json_encode " . json_last_error_msg());
		}
		return $s;
	}

	/**
	 * @param string $form
	 * @return never
	 */
	public function formRedirect($form) {
		//ob_clean();
		echo '<html><head><meta http-equiv="Content-Type" content="text/html; charset=utf-8"/></head><body>';
		echo $form;
		echo "</body></html>";
		//ob_flush();
		//flush();
		UniErr::reset();
		exit();
	}

	/**
	 * @param string $url empty string for redirect onto current page
	 * @param array<string, string|int|float> $data
	 * @return never
	 */
	public function postRedirect($url, $data) {
		echo '<html><head><meta http-equiv="Content-Type" content="text/html; charset=utf-8"/></head><body>';
		echo '<form id="redirectForm" action="' . htmlspecialchars($url) . '" method="post" accept-charset="UTF-8" enctype="application/x-www-form-urlencoded">';
		foreach ($data as $key => $value) {
			echo '<input type="hidden" name="' . htmlspecialchars($key, ENT_QUOTES) . '" value="' . htmlspecialchars((string)$value, ENT_QUOTES) . '">';
		}
		echo '</form>';
		echo '<script type="text/javascript">document.getElementById("redirectForm").submit();</script>';
		echo "</body></html>";
		UniErr::reset();
		exit();
	}

	/**
	 * @param bool $sessionOk
	 * @return never|void
	 */
	public function crossPostFix($sessionOk) {
		if (!$sessionOk && empty($_REQUEST['_uniAntiLoop'])) {
			foreach (headers_list() as $h) { // clear possible new session id header
				header_remove(explode(':', $h)[0]);
			}

			if ($_SERVER['REQUEST_METHOD'] === 'POST') {
				$postData = $_POST + array('_uniAntiLoop'=>1);
				$this->logger->writeLogNoNewLines("INFO UniModul: Recovering lost cross post (POST) cookies ". var_export($postData, true));
				$this->postRedirect("", $postData);
			} else if ($_SERVER['REQUEST_METHOD'] === 'GET') {
				$redirUrl = $this->appendUrlQueryArgs($this->getCurrentRequestUrl(), array('_uniAntiLoop' => '1'));
				$this->logger->writeLogNoNewLines("INFO UniModul: Recovering lost cross post (GET) cookies ". $redirUrl);
				echo '<script type="text/javascript">window.location.href="'.$redirUrl.'";</script>'; // header location redirect won't work -> using js
				exit();
			}
		}
	}

	/**
	 * @param string $redirUrl
	 * @return never
	 */
	public function jsRedirect($redirUrl) { // special function for Google Analytics to forget Referrer
		$script = "<script>setTimeout(function() { window.location.href='".$redirUrl."';}, 200);</script><noscript><a href='".$redirUrl."'>Klikněte prosím/ Please click</a></noscript>";
		$this->formRedirect($script);
	}

	/**
	 * @param string $redirUrl
	 * @return never
	 */
	public function redirect($redirUrl) {
		if (!headers_sent()) {
			header('Location: ' . $redirUrl);
		} else {
			echo '<script type="text/javascript">window.location.href="'.$redirUrl.'";</script>';
		}
		exit();
	}

	/** @return bool  */
	public function isSafariBrowser() {
		if (!isset($_SERVER['HTTP_USER_AGENT'])) return false;
		$agent = (string)$_SERVER['HTTP_USER_AGENT'];
		return strpos($agent, 'Safari') !== false && strpos($agent, 'Chrome') === false;
	}


	/** @var array<string, string> */
	private $isoLangConversionTable = [ "AW" => "ABW", "AF" => "AFG", "AO" => "AGO", "AI" => "AIA", "AX" => "ALA", "AL" => "ALB", "AD" => "AND", "AE" => "ARE", "AR" => "ARG", "AM" => "ARM", "AS" => "ASM", "AQ" => "ATA", "TF" => "ATF", "AG" => "ATG", "AU" => "AUS", "AT" => "AUT", "AZ" => "AZE", "BI" => "BDI", "BE" => "BEL", "BJ" => "BEN", "BQ" => "BES", "BF" => "BFA", "BD" => "BGD", "BG" => "BGR", "BH" => "BHR", "BS" => "BHS", "BA" => "BIH", "BL" => "BLM", "BY" => "BLR", "BZ" => "BLZ", "BM" => "BMU", "BO" => "BOL", "BR" => "BRA", "BB" => "BRB", "BN" => "BRN", "BT" => "BTN", "BV" => "BVT", "BW" => "BWA", "CF" => "CAF", "CA" => "CAN", "CC" => "CCK", "CH" => "CHE", "CL" => "CHL", "CN" => "CHN", "CI" => "CIV", "CM" => "CMR", "CD" => "COD", "CG" => "COG", "CK" => "COK", "CO" => "COL", "KM" => "COM", "CV" => "CPV", "CR" => "CRI", "CU" => "CUB", "CW" => "CUW", "CX" => "CXR", "KY" => "CYM", "CY" => "CYP", "CZ" => "CZE", "DE" => "DEU", "DJ" => "DJI", "DM" => "DMA", "DK" => "DNK", "DO" => "DOM", "DZ" => "DZA", "EC" => "ECU", "EG" => "EGY", "ER" => "ERI", "EH" => "ESH", "ES" => "ESP", "EE" => "EST", "ET" => "ETH", "FI" => "FIN", "FJ" => "FJI", "FK" => "FLK", "FR" => "FRA", "FO" => "FRO", "FM" => "FSM", "GA" => "GAB", "GB" => "GBR", "GE" => "GEO", "GG" => "GGY", "GH" => "GHA", "GI" => "GIB", "GN" => "GIN", "GP" => "GLP", "GM" => "GMB", "GW" => "GNB", "GQ" => "GNQ", "GR" => "GRC", "GD" => "GRD", "GL" => "GRL", "GT" => "GTM", "GF" => "GUF", "GU" => "GUM", "GY" => "GUY", "HK" => "HKG", "HM" => "HMD", "HN" => "HND", "HR" => "HRV", "HT" => "HTI", "HU" => "HUN", "ID" => "IDN", "IM" => "IMN", "IN" => "IND", "IO" => "IOT", "IE" => "IRL", "IR" => "IRN", "IQ" => "IRQ", "IS" => "ISL", "IL" => "ISR", "IT" => "ITA", "JM" => "JAM", "JE" => "JEY", "JO" => "JOR", "JP" => "JPN", "KZ" => "KAZ", "KE" => "KEN", "KG" => "KGZ", "KH" => "KHM", "KI" => "KIR", "KN" => "KNA", "KR" => "KOR", "KW" => "KWT", "LA" => "LAO", "LB" => "LBN", "LR" => "LBR", "LY" => "LBY", "LC" => "LCA", "LI" => "LIE", "LK" => "LKA", "LS" => "LSO", "LT" => "LTU", "LU" => "LUX", "LV" => "LVA", "MO" => "MAC", "MF" => "MAF", "MA" => "MAR", "MC" => "MCO", "MD" => "MDA", "MG" => "MDG", "MV" => "MDV", "MX" => "MEX", "MH" => "MHL", "MK" => "MKD", "ML" => "MLI", "MT" => "MLT", "MM" => "MMR", "ME" => "MNE", "MN" => "MNG", "MP" => "MNP", "MZ" => "MOZ", "MR" => "MRT", "MS" => "MSR", "MQ" => "MTQ", "MU" => "MUS", "MW" => "MWI", "MY" => "MYS", "YT" => "MYT", "NA" => "NAM", "NC" => "NCL", "NE" => "NER", "NF" => "NFK", "NG" => "NGA", "NI" => "NIC", "NU" => "NIU", "NL" => "NLD", "NO" => "NOR", "NP" => "NPL", "NR" => "NRU", "NZ" => "NZL", "OM" => "OMN", "PK" => "PAK", "PA" => "PAN", "PN" => "PCN", "PE" => "PER", "PH" => "PHL", "PW" => "PLW", "PG" => "PNG", "PL" => "POL", "PR" => "PRI", "KP" => "PRK", "PT" => "PRT", "PY" => "PRY", "PS" => "PSE", "PF" => "PYF", "QA" => "QAT", "RE" => "REU", "RO" => "ROU", "RU" => "RUS", "RW" => "RWA", "SA" => "SAU", "SD" => "SDN", "SN" => "SEN", "SG" => "SGP", "GS" => "SGS", "SH" => "SHN", "SJ" => "SJM", "SB" => "SLB", "SL" => "SLE", "SV" => "SLV", "SM" => "SMR", "SO" => "SOM", "PM" => "SPM", "RS" => "SRB", "SS" => "SSD", "ST" => "STP", "SR" => "SUR", "SK" => "SVK", "SI" => "SVN", "SE" => "SWE", "SZ" => "SWZ", "SX" => "SXM", "SC" => "SYC", "SY" => "SYR", "TC" => "TCA", "TD" => "TCD", "TG" => "TGO", "TH" => "THA", "TJ" => "TJK", "TK" => "TKL", "TM" => "TKM", "TL" => "TLS", "TO" => "TON", "TT" => "TTO", "TN" => "TUN", "TR" => "TUR", "TV" => "TUV", "TW" => "TWN", "TZ" => "TZA", "UG" => "UGA", "UA" => "UKR", "UM" => "UMI", "UY" => "URY", "US" => "USA", "UZ" => "UZB", "VA" => "VAT", "VC" => "VCT", "VE" => "VEN", "VG" => "VGB", "VI" => "VIR", "VN" => "VNM", "VU" => "VUT", "WF" => "WLF", "WS" => "WSM", "YE" => "YEM", "ZA" => "ZAF", "ZM" => "ZMB", "ZW" => "ZWE"];

	/**
	 * @param ?string $langCodeAlpha2
	 *
	 * @return ?string
	 */
	public function isoCountryCodeAlpha2ToAplha3($langCodeAlpha2) {
		if (!isset($this->isoLangConversionTable[$langCodeAlpha2])) {
			$this->logger->writeLog("WARN UniModul: Couldn't convert ISO Alpha 2 country code '$langCodeAlpha2'. Code not recognized!");
			return $langCodeAlpha2;
		} else if (!is_null($langCodeAlpha2)) {
			return $this->isoLangConversionTable[$langCodeAlpha2];
		} else {
			return null;
		}
	}


	/** @return string  */
	public static function getCurrentRequestUrl() {
		$pageURL = 'http';
		if (isset($_SERVER["HTTPS"]) && $_SERVER["HTTPS"] == "on") {$pageURL .= "s";}
		$pageURL .= "://";
		$host = Http::server("HTTP_HOST");
		if (isset($_SERVER["HTTP_X_FORWARDED_HOST"])) {
			$host = $_SERVER["HTTP_X_FORWARDED_HOST"];
		}
		$pageURL .= $host . Http::server("REQUEST_URI");
		return $pageURL;
	}

}



class HttpRequestException extends Exception {

	/**
	 * @var int
	 */
	private $status;

	/**
	 * @param int $status
	 * @param string $message
	 * @param int $code
	 * @param Throwable|null $previous
	 */
	public function __construct($status, $message, $code = 0, Throwable $previous = null) {
		$this->status = $status;
        parent::__construct($message, $code, $previous);
    }

    /**
     * @return string
     */
    public function __toString() {
        return __CLASS__ . ": Request returned HTTP Status {$this->status}, and Response: '{$this->message}'\n";
    }

	/**
	 * @return int
	 */
	public function getStatus() {
		return $this->status;
	}
}
